/* 
   glidetv - display TV picture on a 3dfx card and control the TV tuner
   (3dfx card and glide library required)

   Copyright (C) 1997 Ralph Metzler (rjkm@thp.uni-koeln.de)
   Copyright (C) 1998 Heikki Hannikainen <hessu@pspt.fi>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   
   Changelog:
   	Initial version by Ralph Metzler (rjkm@thp.uni-koeln.de)
   	sometime during the year 1997...
   
   	31-Jan-1998	Heikki Hannikainen <hessu@pspt.fi>:

   		- Reformatted (indent...)
   		- Split up the stuff to functions, comment it to some degree
   		  (just to make myself clear how it works...)
   		- Added support for xtvscreen's config file ~/.xtvrc
   		  for input of channel data
   		- Added commands for selecting channels
   		- 7-segment LED styled channel indicator (ugly code)
   
 */

#define CFGLEN 256

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <ctype.h>

#include <glide.h>
#include <linutil.h>

#include "../driver/bttv.h"
#include "../driver/bt848.h"

struct channel_t {
	char *name;
	int freq;
	int nr;
	struct channel_t *next;
	struct channel_t *prev;
} *channels = NULL;

char	*bttvname = "/dev/bttv";	/* BTTV device name */
int	fbttv;				/* File descriptor for the device */
struct video_window vwin;
struct video_buffer vbuf;
int one=1;
int zero=0;
	

/* Real video resolution used: */
int width = 768, height = 576;		/* set this to 640 x 480 for NTSC */

/* Glide graphics: */
GrHwConfiguration hwconfig;		/* Hardware configuration */
GrLfbInfo_t info;

/* LFB array: */
unsigned short *lfb;

void fail(char *msg)
{
  printf(msg);
  grGlideShutdown();
  exit(1);
}

/*
 *	Initialize the Glide library
 */

void init_glide(void)
{
	GrScreenResolution_t resolution = GR_RESOLUTION_800x600;
	
	float scrWidth = 800.0f;
	float scrHeight = 600.0f;
	
	grGlideInit();
	assert(grSstQueryHardware(&hwconfig));
	grSstSelect(0);
	assert(grSstWinOpen(0, resolution, GR_REFRESH_60Hz, GR_COLORFORMAT_ABGR,
			    GR_ORIGIN_UPPER_LEFT, 2, 1));
	grDitherMode(GR_DITHER_DISABLE);
	grClipWindow(0, 0, (FxU32) scrWidth, (FxU32) scrHeight);
	grBufferClear(0x00, 0, 0);
	
	info.size = sizeof(GrLfbInfo_t);
	
	if (grLfbLock(GR_LFB_WRITE_ONLY, GR_BUFFER_FRONTBUFFER,
	       GR_LFBWRITEMODE_565, GR_ORIGIN_UPPER_LEFT, FXFALSE, &info)
	    == FXFALSE) {
		fprintf(stderr, "grLfbLock: Failed to take write lock\n");
		grGlideShutdown();
		exit(1);
	}
	
	lfb = info.lfbPtr;
}

/*
 *	Close the Glide library
 */

void close_glide(void)
{
	grLfbUnlock(GR_LFB_WRITE_ONLY, GR_BUFFER_FRONTBUFFER);
	grGlideShutdown();
}

/*
 *	Drawing on the lfb
 */

void set_bit(int x, int y, unsigned short i)
{
	*(lfb + x + y * 1024) = i;
}

void h_seg(int x, int y, unsigned short val)
{
	int i;
	
	for (i = x+2; i <= x + 9; i++) {
		set_bit(i, y, val);
		set_bit(i, y+1, val);
	}
}

void v_seg(int x, int y, unsigned short val)
{
	int i;
	
	for (i = y+2; i <= y + 11; i++) {
		set_bit(x, i, val);
		set_bit(x+1, i, val);
	}
}

unsigned char segt[10] = { 125, 96, 55, 103, 106, 79, 95, 97, 255, 107 };

void segs(int x, int y, unsigned short val, unsigned char i)
{
	if (i & 1)	h_seg(x, y, val);
	if (i & 2)	h_seg(x, y+12, val);
	if (i & 4)	h_seg(x, y+12+12, val);
	if (i & 8)	v_seg(x, y, val);
	if (i & 16)	v_seg(x, y+12, val);
	if (i & 32)	v_seg(x+10, y, val);
	if (i & 64)	v_seg(x+10, y+12, val);
}

void draw_chan(int i)
{
	segs(0, 0, 0, segt[8]);
	if (i / 10)
		segs(0, 0, 2024, segt[i / 10]);
	segs(15, 0, 0, segt[8]);
	segs(15, 0, 2024, segt[i % 10]);
}

/*
 *	Open the bttv device
 */

void open_bttv(void)
{
	if ((fbttv = open(bttvname, O_RDWR)) < 0) {
		fprintf(stderr, "Cannot open %s: %s\n", bttvname, strerror(errno));
		exit(1);
	}
}

/*
 *	Initialize the bttv device
 */

void init_bttv(void)
{
	if(ioctl(fbttv, VIDIOCCAPTURE, &zero))
	  fail("VIDIOCCAPTURE failed \n");
	
	if(ioctl(fbttv, VIDIOCGFBUF, &vbuf))
	  fail("VIDIOCGFBUF failed \n");
	vbuf.base=info.lfbPtr;
	vbuf.depth=16;
	vbuf.bytesperline=info.strideInBytes;
	vbuf.width=800;
	vbuf.height=600;
	if(ioctl(fbttv, VIDIOCSFBUF, &vbuf))
	  fail("VIDIOCSFBUF failed \n");
	
	vwin.clipcount=0;
	vwin.flags=0;
	vwin.y=(600 - height) / 2;
	vwin.x=(800 - width) / 2 + 15;	/* Window location */
	vwin.width=width;
	vwin.height=height;
	if (ioctl(fbttv, VIDIOCSWIN, &vwin))
	  fail("VIDIOCSWIN failed \n");

	if (ioctl(fbttv, VIDIOCSCHAN, &zero))
	  fail("VIDIOCSCHAN failed \n");
	
	if(ioctl(fbttv, VIDIOCCAPTURE,&one))
	  fail("VIDIOCCAPTURE failed \n");

}

/*
 *	Close the bttv device
 */

void close_bttv(void)
{
        if(ioctl(fbttv, VIDIOCCAPTURE,&zero))
	  fail("VIDIOCCAPTURE failed \n");
	close(fbttv);
}

/*
 *	Read the configuration
 */

void get_xtvrc(char *path, int size)
{
	const char *home = getenv("HOME");
	const char *file = ".xtvrc";
	
	strncpy(path, home, size);
	strncat(path, "/", size);
	strncat(path, file, size);
}

void read_config(void)
{
	FILE *f;
	char fn[1024];
	char s[CFGLEN];
	char *cmd, *par;
	int i = 0;
	struct channel_t *channel = NULL, *prev, **prevp = &channels;
	
	get_xtvrc(fn, 1024);
	
	if (!(f = fopen(fn, "r"))) {
		fprintf(stderr, "Could not open %s for reading: %s\n", fn, strerror(errno));
		exit(1);
	}
	
	while (fgets(s, CFGLEN, f)) {
		cmd = strtok(s, " ");
		par = strtok(NULL, "\n");
		if (!strcasecmp(cmd, "Channel:")) {
			prev = channel;
			channel = malloc(sizeof(struct channel_t));
			*prevp = channel;
			prevp = &channel->next;
			channel->prev = prev;
			i++;
			channel->nr = i;
			channel->name = strdup(par);
		} else if ((!strcasecmp(cmd, "Frequency:")) && (channel)) {
			channel->freq = atoi(par);
			fprintf(stderr, "Channel %d: %s at %d\n", channel->nr, channel->name, channel->freq);
		} 
	}
	
	fclose(f);
	
	if (!channel) {
		fprintf(stderr, "Uh-oh. No channels found! Need to create ~/.xtvrc with xtvscreen first.\n");
		exit(1);
	}
	
	channel->next = channels;
	channel->next->prev = channel;
}

/*
 *	Start it up
 */

void init(void)
{
	read_config();
	open_bttv();
	init_glide();
	init_bttv();
}

/*
 *	Close it down
 */

void closedown(void)
{
	close_bttv();
	close_glide();
	fprintf(stderr, "TV closed.\n");
	exit(0);
}

/*
 *	Change the tuner frequency
 */
 
void change_channel(struct channel_t *channel)
{
	ulong i;
	i = channel->freq;
	
	if (!channel) {
		fprintf(stderr, "Cannot change to a NULL channel!\n");
		closedown();
	}
	
	fprintf(stderr, "Choosing channel %d: %s\r\n", channel->nr, channel->name);
	draw_chan(channel->nr);
	ioctl(fbttv, BTTV_SETFREQ, &i);
}

/*
 *	Find a channel by number
 */
 
struct channel_t *find_channel(struct channel_t *old, int i)
{
	struct channel_t *ch = channels;
	
	do {
		ch = ch->next;
	} while ((ch->nr != i) && (ch != channels));
		
	if (ch->nr == i)
		return ch;
	else
		return old;
}

/*
 *	Main
 */

void main(int argc, char **argv)
{
	char c;
	struct channel_t *channel;
	int capturing = 1;
	
	init();
	
	channel = channels;
	change_channel(channel);
	
	while (1) {
		c = tolower(lin_getch());
		switch (c) {
		case '+':
		case 'n':
		case ' ':
			channel = channel->next;
			change_channel(channel);
			break;
		case '-':
		case 'p':
			channel = channel->prev;
			change_channel(channel);
			break;
		case 'c':
			if (capturing) {
				capturing = 0;
				if(ioctl(fbttv, VIDIOCCAPTURE,&zero))
				  fail("VIDIOCCAPTURE failed \n");
				close_glide();
				fprintf(stderr, "Capturing disabled, press C again to continue...\r\n");
			} else {
				fprintf(stderr, "Capturing enabled.\r\n");
				init_glide();
				if(ioctl(fbttv, VIDIOCCAPTURE,&one))
				  fail("VIDIOCCAPTURE failed \n");
				draw_chan(channel->nr);
				capturing = 1;
			}
			break;
		case 'q':
			closedown();
			break;
		default :
			if ((c >= '1') || (c <= '9')) {
				channel = find_channel(channel, (int)c - 48);
				change_channel(channel);
			}
		}
	}
}

